/************************************************
 * 
 * poolmap.c
 *
 * Multipoint Linkage Disequilibrium Mapping Using Multilocus Allele Frequency Data
 *
 * Implements the method described in the manuscript
 * ``Multipoint Linkage Disequilibrium Mapping Using Multilocus Allele Frequency Data''
 * by Toby Johnson 1,2,3 <toby.johnson@ed.ac.uk>
 *   1 Rothamsted Research, Harpenden, UK
 *   2 Inst. of Evolutionary Biology, University of Edinburgh, Edinburgh, UK
 *   3 Dept. of Zoology, University of British Columbia, Vancouver, CANADA
 * 
 * Published in 
 * Erwin Schr\"odinger Institute preprint series
 *             <http://www.esi.ac.at/Preprint-shadows/esi1433.html>
 *
 * Annals of Human Genetics (in press)
 *
 *
 *
 * This program is Copyright (C) 2002--2004 Toby Johnson <toby.johnson@ed.ac.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * 
 * 
 * 		    GNU GENERAL PUBLIC LICENSE
 * 		       Version 2, June 1991
 * 
 *  Copyright (C) 1989, 1991 Free Software Foundation, Inc.
 *                        59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *  Everyone is permitted to copy and distribute verbatim copies
 *  of this license document, but changing it is not allowed.
 * 
 * 			    Preamble
 * 
 *   The licenses for most software are designed to take away your
 * freedom to share and change it.  By contrast, the GNU General Public
 * License is intended to guarantee your freedom to share and change free
 * software--to make sure the software is free for all its users.  This
 * General Public License applies to most of the Free Software
 * Foundation's software and to any other program whose authors commit to
 * using it.  (Some other Free Software Foundation software is covered by
 * the GNU Library General Public License instead.)  You can apply it to
 * your programs, too.
 * 
 *   When we speak of free software, we are referring to freedom, not
 * price.  Our General Public Licenses are designed to make sure that you
 * have the freedom to distribute copies of free software (and charge for
 * this service if you wish), that you receive source code or can get it
 * if you want it, that you can change the software or use pieces of it
 * in new free programs; and that you know you can do these things.
 * 
 *   To protect your rights, we need to make restrictions that forbid
 * anyone to deny you these rights or to ask you to surrender the rights.
 * These restrictions translate to certain responsibilities for you if you
 * distribute copies of the software, or if you modify it.
 * 
 *   For example, if you distribute copies of such a program, whether
 * gratis or for a fee, you must give the recipients all the rights that
 * you have.  You must make sure that they, too, receive or can get the
 * source code.  And you must show them these terms so they know their
 * rights.
 * 
 *   We protect your rights with two steps: (1) copyright the software, and
 * (2) offer you this license which gives you legal permission to copy,
 * distribute and/or modify the software.
 * 
 *   Also, for each author's protection and ours, we want to make certain
 * that everyone understands that there is no warranty for this free
 * software.  If the software is modified by someone else and passed on, we
 * want its recipients to know that what they have is not the original, so
 * that any problems introduced by others will not reflect on the original
 * authors' reputations.
 * 
 *   Finally, any free program is threatened constantly by software
 * patents.  We wish to avoid the danger that redistributors of a free
 * program will individually obtain patent licenses, in effect making the
 * program proprietary.  To prevent this, we have made it clear that any
 * patent must be licensed for everyone's free use or not licensed at all.
 * 
 *   The precise terms and conditions for copying, distribution and
 * modification follow.
 * 
 * 		    GNU GENERAL PUBLIC LICENSE
 *    TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION
 * 
 *   0. This License applies to any program or other work which contains
 * a notice placed by the copyright holder saying it may be distributed
 * under the terms of this General Public License.  The "Program", below,
 * refers to any such program or work, and a "work based on the Program"
 * means either the Program or any derivative work under copyright law:
 * that is to say, a work containing the Program or a portion of it,
 * either verbatim or with modifications and/or translated into another
 * language.  (Hereinafter, translation is included without limitation in
 * the term "modification".)  Each licensee is addressed as "you".
 * 
 * Activities other than copying, distribution and modification are not
 * covered by this License; they are outside its scope.  The act of
 * running the Program is not restricted, and the output from the Program
 * is covered only if its contents constitute a work based on the
 * Program (independent of having been made by running the Program).
 * Whether that is true depends on what the Program does.
 * 
 *   1. You may copy and distribute verbatim copies of the Program's
 * source code as you receive it, in any medium, provided that you
 * conspicuously and appropriately publish on each copy an appropriate
 * copyright notice and disclaimer of warranty; keep intact all the
 * notices that refer to this License and to the absence of any warranty;
 * and give any other recipients of the Program a copy of this License
 * along with the Program.
 * 
 * You may charge a fee for the physical act of transferring a copy, and
 * you may at your option offer warranty protection in exchange for a fee.
 * 
 *   2. You may modify your copy or copies of the Program or any portion
 * of it, thus forming a work based on the Program, and copy and
 * distribute such modifications or work under the terms of Section 1
 * above, provided that you also meet all of these conditions:
 * 
 *     a) You must cause the modified files to carry prominent notices
 *     stating that you changed the files and the date of any change.
 * 
 *     b) You must cause any work that you distribute or publish, that in
 *     whole or in part contains or is derived from the Program or any
 *     part thereof, to be licensed as a whole at no charge to all third
 *     parties under the terms of this License.
 * 
 *     c) If the modified program normally reads commands interactively
 *     when run, you must cause it, when started running for such
 *     interactive use in the most ordinary way, to print or display an
 *     announcement including an appropriate copyright notice and a
 *     notice that there is no warranty (or else, saying that you provide
 *     a warranty) and that users may redistribute the program under
 *     these conditions, and telling the user how to view a copy of this
 *     License.  (Exception: if the Program itself is interactive but
 *     does not normally print such an announcement, your work based on
 *     the Program is not required to print an announcement.)
 * 
 * These requirements apply to the modified work as a whole.  If
 * identifiable sections of that work are not derived from the Program,
 * and can be reasonably considered independent and separate works in
 * themselves, then this License, and its terms, do not apply to those
 * sections when you distribute them as separate works.  But when you
 * distribute the same sections as part of a whole which is a work based
 * on the Program, the distribution of the whole must be on the terms of
 * this License, whose permissions for other licensees extend to the
 * entire whole, and thus to each and every part regardless of who wrote it.
 * 
 * Thus, it is not the intent of this section to claim rights or contest
 * your rights to work written entirely by you; rather, the intent is to
 * exercise the right to control the distribution of derivative or
 * collective works based on the Program.
 * 
 * In addition, mere aggregation of another work not based on the Program
 * with the Program (or with a work based on the Program) on a volume of
 * a storage or distribution medium does not bring the other work under
 * the scope of this License.
 * 
 *   3. You may copy and distribute the Program (or a work based on it,
 * under Section 2) in object code or executable form under the terms of
 * Sections 1 and 2 above provided that you also do one of the following:
 * 
 *     a) Accompany it with the complete corresponding machine-readable
 *     source code, which must be distributed under the terms of Sections
 *     1 and 2 above on a medium customarily used for software interchange; or,
 * 
 *     b) Accompany it with a written offer, valid for at least three
 *     years, to give any third party, for a charge no more than your
 *     cost of physically performing source distribution, a complete
 *     machine-readable copy of the corresponding source code, to be
 *     distributed under the terms of Sections 1 and 2 above on a medium
 *     customarily used for software interchange; or,
 * 
 *     c) Accompany it with the information you received as to the offer
 *     to distribute corresponding source code.  (This alternative is
 *     allowed only for noncommercial distribution and only if you
 *     received the program in object code or executable form with such
 *     an offer, in accord with Subsection b above.)
 * 
 * The source code for a work means the preferred form of the work for
 * making modifications to it.  For an executable work, complete source
 * code means all the source code for all modules it contains, plus any
 * associated interface definition files, plus the scripts used to
 * control compilation and installation of the executable.  However, as a
 * special exception, the source code distributed need not include
 * anything that is normally distributed (in either source or binary
 * form) with the major components (compiler, kernel, and so on) of the
 * operating system on which the executable runs, unless that component
 * itself accompanies the executable.
 * 
 * If distribution of executable or object code is made by offering
 * access to copy from a designated place, then offering equivalent
 * access to copy the source code from the same place counts as
 * distribution of the source code, even though third parties are not
 * compelled to copy the source along with the object code.
 * 
 *   4. You may not copy, modify, sublicense, or distribute the Program
 * except as expressly provided under this License.  Any attempt
 * otherwise to copy, modify, sublicense or distribute the Program is
 * void, and will automatically terminate your rights under this License.
 * However, parties who have received copies, or rights, from you under
 * this License will not have their licenses terminated so long as such
 * parties remain in full compliance.
 * 
 *   5. You are not required to accept this License, since you have not
 * signed it.  However, nothing else grants you permission to modify or
 * distribute the Program or its derivative works.  These actions are
 * prohibited by law if you do not accept this License.  Therefore, by
 * modifying or distributing the Program (or any work based on the
 * Program), you indicate your acceptance of this License to do so, and
 * all its terms and conditions for copying, distributing or modifying
 * the Program or works based on it.
 * 
 *   6. Each time you redistribute the Program (or any work based on the
 * Program), the recipient automatically receives a license from the
 * original licensor to copy, distribute or modify the Program subject to
 * these terms and conditions.  You may not impose any further
 * restrictions on the recipients' exercise of the rights granted herein.
 * You are not responsible for enforcing compliance by third parties to
 * this License.
 * 
 *   7. If, as a consequence of a court judgment or allegation of patent
 * infringement or for any other reason (not limited to patent issues),
 * conditions are imposed on you (whether by court order, agreement or
 * otherwise) that contradict the conditions of this License, they do not
 * excuse you from the conditions of this License.  If you cannot
 * distribute so as to satisfy simultaneously your obligations under this
 * License and any other pertinent obligations, then as a consequence you
 * may not distribute the Program at all.  For example, if a patent
 * license would not permit royalty-free redistribution of the Program by
 * all those who receive copies directly or indirectly through you, then
 * the only way you could satisfy both it and this License would be to
 * refrain entirely from distribution of the Program.
 * 
 * If any portion of this section is held invalid or unenforceable under
 * any particular circumstance, the balance of the section is intended to
 * apply and the section as a whole is intended to apply in other
 * circumstances.
 * 
 * It is not the purpose of this section to induce you to infringe any
 * patents or other property right claims or to contest validity of any
 * such claims; this section has the sole purpose of protecting the
 * integrity of the free software distribution system, which is
 * implemented by public license practices.  Many people have made
 * generous contributions to the wide range of software distributed
 * through that system in reliance on consistent application of that
 * system; it is up to the author/donor to decide if he or she is willing
 * to distribute software through any other system and a licensee cannot
 * impose that choice.
 * 
 * This section is intended to make thoroughly clear what is believed to
 * be a consequence of the rest of this License.
 * 
 *   8. If the distribution and/or use of the Program is restricted in
 * certain countries either by patents or by copyrighted interfaces, the
 * original copyright holder who places the Program under this License
 * may add an explicit geographical distribution limitation excluding
 * those countries, so that distribution is permitted only in or among
 * countries not thus excluded.  In such case, this License incorporates
 * the limitation as if written in the body of this License.
 * 
 *   9. The Free Software Foundation may publish revised and/or new versions
 * of the General Public License from time to time.  Such new versions will
 * be similar in spirit to the present version, but may differ in detail to
 * address new problems or concerns.
 * 
 * Each version is given a distinguishing version number.  If the Program
 * specifies a version number of this License which applies to it and "any
 * later version", you have the option of following the terms and conditions
 * either of that version or of any later version published by the Free
 * Software Foundation.  If the Program does not specify a version number of
 * this License, you may choose any version ever published by the Free Software
 * Foundation.
 * 
 *   10. If you wish to incorporate parts of the Program into other free
 * programs whose distribution conditions are different, write to the author
 * to ask for permission.  For software which is copyrighted by the Free
 * Software Foundation, write to the Free Software Foundation; we sometimes
 * make exceptions for this.  Our decision will be guided by the two goals
 * of preserving the free status of all derivatives of our free software and
 * of promoting the sharing and reuse of software generally.
 * 
 * 			    NO WARRANTY
 * 
 *   11. BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
 * FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT WHEN
 * OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
 * PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED
 * OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS
 * TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE
 * PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING,
 * REPAIR OR CORRECTION.
 * 
 *   12. IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
 * REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES,
 * INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING
 * OUT OF THE USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED
 * TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY
 * YOU OR THIRD PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER
 * PROGRAMS), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 * 
 * 		     END OF TERMS AND CONDITIONS
 * 
 * 
 * 
 *
 *
 *
 ************************************************/


#include <math.h>

#define ILOCUS map[index]
/* a definition used in make_ml_isotone() */

/* function prototypes */
double calc_log_lik(int num_loci, int* pd, int* qd,int* pc, int* qc, int *x,int locus1,int locus2);
double make_ml_isotone (int num_loci, int* pd, int* qd,int* pc, int* qc, int *x,int low,int high);
double lnfact(int n);
int poolmap (int* num_loci, int* pd, int* qd, int* pc, int* qc, double* log_lik_null, double* log_lik_mle, double* log_lik, int* x_mle);


int poolmap (int* num_loci, int* pd, int* qd, int* pc, int* qc, double* log_lik_null, double* log_lik_mle, double* log_lik, int* x_mle) {

  /* pd, qd, pc and qc are counts of zero (p) and one (q) alleles in
     the d(iseased)s and c(ontrol)s */

  /* log_lik has dimension num_loci+1 */

  /* a component of the log-likelihood that depends only on pc and qc */
  double log_lik_offset;

  int loc, loc2; // iterators to count through loci

  int* x; // estimated amount of ancestral chromosome at each locus



  // calculate a constant term of log-lik left out of other calculations ...
  log_lik_offset = 0;
  for (loc = 0;loc < *num_loci;loc++) {
    log_lik_offset += lnfact(pc[loc]+qc[loc]) - lnfact(pc[loc]) - lnfact(qc[loc]);
  }

  x = (int *)malloc(*num_loci * sizeof(int));

  /* calculate log-lik of x=0 everywhere */
  for (loc = 0;loc < *num_loci;loc++) {
    x[loc] = 0;
  }
  *log_lik_null = calc_log_lik(*num_loci,pd,qd,pc,qc,x,0,*num_loci-1);

  // calculate profile likelihood

  /* cycle through every possible loc */
  for (loc = 0;loc <= *num_loci;loc++) {

    // find MLE(x) under monotonicity restriction and calculate log_lik
    log_lik[loc] = log_lik_offset;
    if (loc > 0) {
      log_lik[loc] += make_ml_isotone(*num_loci,pd,qd,pc,qc,x,0,loc-1);
    }
    if (loc < *num_loci) {
      log_lik[loc] += make_ml_isotone(*num_loci,pd,qd,pc,qc,x,*num_loci-1,loc);
    }
    
    // if highest log_lik so far, update MLE(x) and log_lik_mle
    if (loc == 0 || log_lik[loc] > *log_lik_mle) {
      for (loc2 = 0;loc2 < *num_loci; loc2++) {
	x_mle[loc2] = x[loc2];
      }
      *log_lik_mle = log_lik[loc];
    }
    
  } // end for loop over loc

  free(x);
  
  return(0);
}

  


double calc_log_lik(int num_loci, int* pd, int* qd,int* pc, int* qc, int *x,int locus1,int locus2) {


  int locus;
  /* counter */
  int start;
  int end;
  /* set from locus1 and locus2 so that start<=end */

  double log_lik;
  /* cumulative */

  int total_p;
  int total_q;
  /* counts of 0 and 1 for   d(iseased)s + c(ontrols) - a(ncestral)s */

  double freq_p;
  double freq_q;
  /* (R)MLE allele frequencies */

  double min_p,max_p;

  double coeff_p;
  double coeff_q;

  double best_p;
  double best_log_lik;
  double trial_p;
  double trial_log_lik;

  if (locus1 <= locus2) {
    start = locus1;
    end = locus2;
  } else {
    start = locus2;
    end = locus1;
  }
  
  log_lik = 0;

  for (locus=start;locus<=end;locus++) {

    total_p = pd[locus]+pc[locus];
    total_q = qd[locus]+qc[locus];

    //  if (x[locus] == 0) {
    if (0) {
      /* x==0 has analytic solution */
      freq_p = (double)total_p / (total_p+total_q);

      freq_q = 1. - freq_p;
      log_lik += total_p * log(freq_p) + total_q * log(freq_q) + 
	lnfact(pd[locus]+qd[locus]) - lnfact(pd[locus]) - lnfact(qd[locus]);

    } else {
      
      if (x[locus] > pd[locus]) {
	/* only a=1 is possible, analytic solution */
	freq_p = (double)(total_p)/(1+total_p+total_q-x[locus]);
	freq_q = 1. - freq_p;
      	log_lik += total_p * log(freq_p) + (total_q+1-x[locus]) * log(freq_q) + 
	  lnfact(pd[locus]+qd[locus]-x[locus]) - lnfact(pd[locus]) - lnfact(qd[locus]-x[locus]);
	
      } else {
	if (x[locus] > qd[locus]) {
	  /* only a=0 is possible, analytic solution */
	  freq_p = (double)(1+total_p-x[locus])/(1+total_p+total_q-x[locus]);
	  freq_q = 1. - freq_p;
	  log_lik += (total_p+1-x[locus]) * log(freq_p) + total_q * log(freq_q) + 
      	    lnfact(pd[locus]-x[locus]+qd[locus]) - lnfact(pd[locus]-x[locus]) - lnfact(qd[locus]);
	  
	  
	  
	} else {

	  /* need to sum over a=0 and a=1 */

	  if (x[locus] > pd[locus]) {
	    coeff_p = 0;
	  } else {
	    coeff_p = exp(lnfact(pd[locus]+qd[locus]-x[locus]) - 
			lnfact(pd[locus]-x[locus]) - lnfact(qd[locus]));
	  }

	  if (x[locus] > qd[locus]) {
	    coeff_q = 0;
	  } else {
	    coeff_q = exp(lnfact(pd[locus]+qd[locus]-x[locus]) - 
			lnfact(pd[locus]) - lnfact(qd[locus]-x[locus]));
	  }

	  /* log-lik = total_p * log(p) + total_q * log(q) +
	     log(p coeff_p pow(p,-x) + (1-p) coeff_q pow(1-p,-x)) 
	     CAN BE MULTIMODAL, so do a simple grid search */
	  
	  // if sum over a is trivial, can do analytically
	  // otherwise can BOUND the region to be searched ...
	  
	  min_p = (double)(1+total_p-x[locus])/(1+total_p+total_q-x[locus]);
	  max_p = (double)(total_p)/(1+total_p+total_q-x[locus]);
	  
	  best_p = min_p;
	  best_log_lik = total_p * log(best_p) + total_q * log(1-best_p) +
	    log(coeff_p * pow(best_p,1-x[locus]) + coeff_q * pow(1-best_p,1-x[locus]));
	  
	  for (trial_p = min_p+0.01; trial_p <= max_p; trial_p += 0.01) {
	    trial_log_lik = total_p * log(trial_p) + total_q * log(1-trial_p) +
	      log(coeff_p * pow(trial_p,1-x[locus]) + coeff_q * pow(1-trial_p,1-x[locus]));
	    
	    if (trial_log_lik > best_log_lik) {
	      best_log_lik = trial_log_lik;
	      best_p = trial_p;
	    }
	  }
	  

	  freq_p = best_p;
	  freq_q = 1-freq_p;
	  
	  log_lik += total_p * log(freq_p) + total_q * log(freq_q) + 
	    log(coeff_p * pow(freq_p,1-x[locus]) + coeff_q * pow(freq_q,1-x[locus]));
	}
      }
      
    }
  }
  
  return (log_lik);
} /* calc_log_lik */



/********************
 *                  *			  
 *  make_ml_isotone *
 *                  *			  
 ********************/



double make_ml_isotone (int num_loci, int* pd, int* qd,int* pc, int* qc, int *x,int low,int high) {
  /* make_ml_isotone
     this is the guts of the program

     more documentation is needed


  */

  int index;
  /* counter, maps to locus */
  int istart;
  int iend;
  /* limits for index counter */

  int* map = (int *)malloc(num_loci*sizeof(int)); /* locus = map[index] && index = map[locus] ALWAYS */

  int* xmap = (int *)malloc(num_loci*sizeof(int)); /* a local copy of x[map[index]] */

  int* x_max = (int *)malloc(num_loci*sizeof(int)); /* maximum a for a given locus = pa (x=0) or qa (x=1) */

  double best_log_lik;
  double this_log_lik;
  double last_log_lik; //
  int best_x; //          these two only used for paranoid checking mode ...

  double total_log_lik;

  int av_flag; // flag set if adjacent violators exist

  int in_pool;
  int pool_start;
  int pool_end;

  int pool_x_max;


  /* detect which direction we're going in */

  if (low<=high) {
    istart = low;
    iend = high;
    for (index=istart;index<=iend;index++) {
      map[index] = index;
    }
  } else {
    istart = high;
    iend = low;
    for (index=istart;index<=iend;index++) {
      map[index] = high+low-index;
    }
  }

  // set up x_max 

  for (index=istart;index<=iend;index++) {
    if (pd[ILOCUS] > qd[ILOCUS]) {
      x_max[ILOCUS] = pd[ILOCUS];
    } else {
      x_max[ILOCUS] = qd[ILOCUS];
    }
    //    if (x_max[ILOCUS] > glob_x_max) {
    //  x_max[ILOCUS] = glob_x_max;
    //}
  }

  // make non-isotonic mle //

  for (index=istart;index<=iend;index++) {
    x[ILOCUS] = 0;
    best_log_lik = calc_log_lik(num_loci,pd,qd,pc,qc,x,ILOCUS,ILOCUS);
    best_x = 0;
    
    for(x[ILOCUS]=1;x[ILOCUS]<=x_max[ILOCUS];x[ILOCUS]++) {
      this_log_lik = calc_log_lik(num_loci,pd,qd,pc,qc,x,ILOCUS,ILOCUS);
      
      if (this_log_lik > best_log_lik) {
	best_log_lik = this_log_lik;
	best_x = x[ILOCUS];
      }
    }

    x[ILOCUS] = best_x;

  }

  // if only one locus, no need to do PAVA
  if (istart == iend) {
    // calculate total_log_lik
    total_log_lik = calc_log_lik(num_loci,pd,qd,pc,qc,x,low,high);

    return(total_log_lik);
  }

  // otherwise look for adjacent violators 
  
  // initialise xmap
  for (index=istart;index<=iend;index++) {
    xmap[index] = x[map[index]];
  }

  do {
    av_flag = 0;

    for (index=istart;index<=iend-1;index++) {
      
      pool_start=index;
      // can we start a pool here? if not (a==-1) cycle ...
      if (xmap[pool_start]!=-1) {
	
	// LOOK FOR A POOL
	for (pool_end=pool_start+1;pool_end<=iend;pool_end++) {
	  // run over all -1s looking to see the next non -1 a 
	  if (xmap[pool_end] != -1) {
	    if (xmap[pool_end] >= xmap[pool_start]) {
	      pool_end = pool_start;
	      // this makes the pool size 1, so nothing happens
	    } else {
	      // otherwise we've got a pool size >= 2 so will PAV
	      xmap[pool_end] = -1;
	    }
	    //  in either case:
	    break;
	  }
	}

	if (pool_end>iend) {
	  // we reached the end of the interval and never found a pool
	  pool_end = pool_start;
	} 


	// POOL ADJACENT VIOLATORS ...
	if (pool_end != pool_start) {
	  av_flag = 1;
	  // have found a pool size >= 2 ! see comments above

	  // check whether the pool_end found is followed by -1's, in which 
	  // case include them in the pool
	  while (pool_end < iend && xmap[pool_end+1]==-1) {
	    pool_end ++;
	  }
	  
	  
	  
	  pool_x_max = x_max[map[pool_start]];
	  for (in_pool=pool_start+1;in_pool<=pool_end;in_pool++) {
	    if (x_max[map[in_pool]] < pool_x_max) {
	      pool_x_max = x_max[map[in_pool]];
	    }
	  }
	  
	  
	  for (in_pool=pool_start;in_pool<=pool_end;in_pool++) {
	    x[map[in_pool]] = 0; // crucial set all a's equal
	  }
	  best_log_lik = calc_log_lik(num_loci,pd,qd,pc,qc,x,map[pool_start],map[pool_end]);
	  best_x = 0;

	  for(xmap[pool_start]=1;xmap[pool_start]<=pool_x_max;xmap[pool_start]++) {
	    
	    for (in_pool=pool_start;in_pool<=pool_end;in_pool++) {
	      x[map[in_pool]] = xmap[pool_start]; // crucial set all a's equal
	    }
	    
	    this_log_lik = calc_log_lik(num_loci,pd,qd,pc,qc,x,map[pool_start],map[pool_end]);
	    if (this_log_lik > best_log_lik) {
	      best_log_lik = this_log_lik;
	      best_x = xmap[pool_start];
	    }
	  }

	  xmap[pool_start] = best_x;
	  for (in_pool=pool_start;in_pool<=pool_end;in_pool++) {
	    x[map[in_pool]] = best_x; // crucial set all a's equal
	  }

	  break;// the for loop of pool_start hunting
	}
	
      } // endif (x[map[pool_start]]!=-1)
      
    } // end for
  } while (av_flag);

  free(x_max);
  free(xmap);
  free(map);
  
  // calculate total_log_lik
  total_log_lik = calc_log_lik(num_loci,pd,qd,pc,qc,x,low,high);
  
  return(total_log_lik);
} /* make_ml_isotone */
  




double lnfact(int n) {
  /* returns ln(n!), using a lookup table for 0<=n<=29 and the glibc
     lgamma() function otherwise.  if your glibc does not have the
     lgamma function, you could use gammln from Numerical Recipes, or
     something from the GSL*/

  static double lnfactcache[30] = {0.000000,0.000000,0.693147,1.791759,3.178054,
				   4.787492,6.579251,8.525161,10.604603,12.801827,
		 		   15.104413,17.502308,19.987214,22.552164,25.191221,
			 	   27.899271,30.671860,33.505073,36.395445,39.339884,
				   42.335616,45.380139,48.471181,51.606676,54.784729,
				   58.003605,61.261702,64.557539,67.889743,71.257039};

  if (0 <= n && n < 30) {
    return (lnfactcache[n]);
  }

  return (lgamma((double)(n+1)));
}

